
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <libusb-1.0/libusb.h>

#include <iostream>
#include <fstream>
#include <string>
#include <map>
#include <algorithm>

using namespace std;

// Values for bmRequestType in the Setup transaction's Data packet.

static const int CONTROL_REQUEST_TYPE_IN = LIBUSB_ENDPOINT_IN | LIBUSB_REQUEST_TYPE_CLASS | LIBUSB_RECIPIENT_INTERFACE;

// From the HID spec:

static const int HID_GET_REPORT = 0x01;
static const int HID_REPORT_TYPE_FEATURE = 0x03;

static const int INTERFACE_NUMBER = 0;
static const int TIMEOUT_MS = 500;

int got_signal = 0;

void handler(int signal)
{
    got_signal = signal;
}

static const char *usberror(int ret)
{
  switch(ret) {
    case LIBUSB_SUCCESS:
      return "Success (no error).";

    case LIBUSB_ERROR_IO:
      return "Input/output error.";

    case LIBUSB_ERROR_INVALID_PARAM: 	
      return "Invalid parameter.";

    case LIBUSB_ERROR_ACCESS:
      return "Access denied (insufficient permissions).";

    case LIBUSB_ERROR_NO_DEVICE:
      return "No such device (it may have been disconnected).";

    case LIBUSB_ERROR_NOT_FOUND:
      return "Entity not found.";

    case LIBUSB_ERROR_BUSY:
      return "Resource busy.";

    case LIBUSB_ERROR_TIMEOUT:
      return "Operation timed out.";

    case LIBUSB_ERROR_OVERFLOW:
      return "Overflow.";

    case LIBUSB_ERROR_PIPE: 	
      return "Pipe error.";

    case LIBUSB_ERROR_INTERRUPTED:
      return "System call interrupted (perhaps due to signal).";

    case LIBUSB_ERROR_NO_MEM:
      return "Insufficient memory.";

    case LIBUSB_ERROR_NOT_SUPPORTED:
      return "Operation not supported or unimplemented on this platform.";

    case LIBUSB_ERROR_OTHER:
      return "Other error. ";
    }
    return "unknown error";
}

int queryVersion(libusb_device_handle *devh)
{
    static const int MAX_CONTROL_IN_TRANSFER_SIZE = 64;

	int bytes_received;
	unsigned char data_in[MAX_CONTROL_IN_TRANSFER_SIZE];
	int i = 0;
	int result = 0;

	bytes_received = libusb_control_transfer(
			devh,
			CONTROL_REQUEST_TYPE_IN ,
			HID_GET_REPORT,
		    (HID_REPORT_TYPE_FEATURE<<8)|0x07,
			INTERFACE_NUMBER,
			data_in,
			MAX_CONTROL_IN_TRANSFER_SIZE,
			TIMEOUT_MS);

	if (bytes_received >= 0)
	{
 		printf("Found USB IR Remote Receiver with IRMP release ");
	 	for(i = 0; i < bytes_received; i++)
	 	{
	 		printf("%c",data_in[i]);
	 	}
		printf("\n");
	}
	else
	{
		fprintf(stderr, "Error receiving IRMP release %s (%d)\n", usberror(result), result);
		return result;
	}
	return 0;
 }

struct IRMP_DATA
{
  int               protocol;                                               // protocol, i.e. NEC_PROTOCOL
  int              address;                                                // address
  int              command;                                                // command
  int               flags;                                                  // flags, e.g. repetition

  IRMP_DATA() {
    protocol = 0;
    address = 0;
    command = 0;
    flags = 0;
  }

  bool operator < (const IRMP_DATA& x) const {
    return ( protocol < x.protocol ) | ( address < x.address ) | ( command < x.command );
  }
  bool operator == (const IRMP_DATA& x) const {
    return ( protocol == x.protocol ) & ( address == x.address ) & ( command == x.command );
  }
  IRMP_DATA& operator = (const IRMP_DATA& x) {
    protocol = x.protocol;
    address = x.address;
    command = x.command;
    return *this;
  }
};

map<IRMP_DATA, string> cfg;

int queryIR(libusb_device_handle *devh)
{

	static const int INTERRUPT_IN_ENDPOINT = 0x81;
	static const int MAX_INTERRUPT_IN_TRANSFER_SIZE = 64;

	int bytes_transferred;
	//int i = 0;
	int result = 0;

 	unsigned char data_in[MAX_INTERRUPT_IN_TRANSFER_SIZE];
    IRMP_DATA last;

    do {
	    result = libusb_interrupt_transfer(
			devh,
			INTERRUPT_IN_ENDPOINT,
			data_in,
			MAX_INTERRUPT_IN_TRANSFER_SIZE,
			&bytes_transferred,
			TIMEOUT_MS);

	    if (result >= 0) {
/*
	      	printf("Data received via interrupt transfer:\n");
	      	for(i = 0; i < bytes_transferred; i++)
	      	{
	      		printf("%02x ",(int)data_in[i]);
	      	}
	      	printf("\n");
*/
		    if (bytes_transferred >= 7 && data_in[0] == 1)
		    {
                IRMP_DATA irmp;

                irmp.protocol = data_in[1];
                irmp.address = (data_in[3] << 8) | data_in[2];
                irmp.command = (data_in[5] << 8) | data_in[4];
                irmp.flags = data_in[6];
                bool bFound = false;
                for (map<IRMP_DATA, string>::const_iterator i = cfg.begin(); i != cfg.end(); ++i ) {
                  const IRMP_DATA& c = (*i).first;
                  if(irmp == c) {
                        bFound = true;
                        if(!(last == irmp)) {
                            last = irmp;
                            int ret = system((*i).second.c_str());               
                            if (WIFSIGNALED(ret) && (WTERMSIG(ret) == SIGINT || WTERMSIG(ret) == SIGQUIT)) {
                            }
                        }
                        break;
                    }
                }
                if(!bFound)
    		      	printf("Unhandled data received, protocol %d, address %d, command %d, flags %d\n", irmp.protocol, irmp.address, irmp.command, irmp.flags);
		    }
        } else {
            if(result != LIBUSB_ERROR_TIMEOUT) {
		        fprintf(stderr, "No data received in interrupt transfer %s (%d)\n", usberror(result), result);
		        return -1;
            }
            last.protocol = 0;
	    }
	} while (( got_signal == 0) && (result >= 0 || result == LIBUSB_ERROR_TIMEOUT));

  	return 0;
 }



string strip_space( const string & s ) {
   string ret(s);
   while ( ret.length() && ( ret[0]== ' ' || ret[0]== '\t' )) 
     ret = ret.substr(1);
   while ( ret.length() && ( ret[ret.length()-1]== ' ' || ret[ret.length()-1]== '\t' )) 
     ret = ret.substr(0,ret.length()-1);
   return ret;
}

int parse(const char* szFile) {

   ifstream in(szFile);
   if(!in.good()) {
    return -1;
   }
   string line;
   while ( in.good() && !in.eof() ) {
      getline(in, line);
      line = strip_space(line);
      if ( line.length() && line[0] != '#' ) {
         // cout << line << endl;
	     size_t pos = line.find(":");
	     if ( pos != string::npos ) {
            IRMP_DATA irmp;
    	    string key = strip_space(line.substr(0,pos));

            replace(key.begin(), key.end(), ',', ' ');
            sscanf(key.c_str(),"%d %d %d", &irmp.protocol, &irmp.address, &irmp.command );
    	    string value = strip_space(line.substr(pos+1));
    	    cfg.insert(make_pair(irmp,value));
    	 }
      }
   }
/*
   for ( map<IRMP_DATA, string>::const_iterator i = cfg.begin(); i != cfg.end(); ++i ) {
     const IRMP_DATA& irmp = (*i).first;
     cout << (int) irmp.protocol << "," << (int)irmp.address << "," << (int)irmp.command << " : " << (*i).second << "\n";  
    }
*/
    return cfg.size();
}

int main(int argc, char *argv[])
{
	static const int VENDOR_ID = 0x16c0;
	static const int PRODUCT_ID = 0x05df;

	struct libusb_device_handle *devh = NULL;
	int device_ready = 0;
	int result;
    if(argc != 2) {
	    fprintf(stderr, "Usage %s CONFIG.CFG\n", argv[0]);
        return -1;
    }

    if(parse(argv[1])<=0) {
	    fprintf(stderr, "Can't parse %s\n", argv[1]);
        return -1;
    }

    /* now install our own signal handler */
    signal(SIGHUP, handler);
    signal(SIGINT, handler);
    signal(SIGQUIT, handler);
    signal(SIGTERM, handler);

	result = libusb_init(NULL);
	if (result >= 0)
	{
		devh = libusb_open_device_with_vid_pid(NULL, VENDOR_ID, PRODUCT_ID);
		if (devh != NULL)
		{
			// The HID has been detected.
			// Detach the hidusb driver from the HID to enable using libusb.

			libusb_detach_kernel_driver(devh, INTERFACE_NUMBER);
			{
				result = libusb_claim_interface(devh, INTERFACE_NUMBER);
				if (result >= 0)
				{
					device_ready = 1;
				}
				else
				{
					fprintf(stderr, "libusb_claim_interface error %s (%d)\n", usberror(result), result);
				}
			}
		}
		else
		{
			fprintf(stderr, "Unable to find the device.\n");
		}
	}
	else
	{
		fprintf(stderr, "Unable to initialize libusb %s (%d)\n", usberror(result), result);
	}

	if (device_ready)
	{
		// Send and receive data.

		if( 0 == queryVersion(devh))
    		queryIR(devh);


		// Finished using the device.

		libusb_release_interface(devh, 0);
	}
	libusb_close(devh);
	libusb_exit(NULL);
	return 0;
}
